<?php
/**
 * FuzeWorks Component.
 *
 * The FuzeWorks PHP FrameWork
 *
 * Copyright (C) 2013-2019 TechFuze
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * @author    TechFuze
 * @copyright Copyright (c) 2013 - 2019, TechFuze. (http://techfuze.net)
 * @license   https://opensource.org/licenses/MIT MIT License
 *
 * @link  http://techfuze.net/fuzeworks
 * @since Version 1.2.0
 *
 * @version Version 1.2.0
 */

namespace FuzeWorks\Async\Constraint;
use FuzeWorks\Async\Constraint;
use FuzeWorks\Async\Task;
use FuzeWorks\Async\Tasks;
use FuzeWorks\Async\TasksException;
use FuzeWorks\Exception\FactoryException;
use FuzeWorks\Exception\LibraryException;
use FuzeWorks\Factory;
use FuzeWorks\Libraries;

class GroupDependencyConstraint implements Constraint
{

    /**
     * @var string
     */
    protected $groupName;

    protected $returnStatus = Task::DELAYED;

    public function __construct(string $groupName)
    {
        $this->groupName = $groupName;
    }

    /**
     * @inheritDoc
     */
    public function intervene(Task $task, array $tasks): bool
    {
        // Find whether any task is not completed
        foreach ($tasks as $t)
        {
            foreach ($t->getConstraints() as $constraint)
            {
                // Check whether the constraint is a GroupConstraint
                if ($constraint instanceof GroupConstraint && $this->groupName === $constraint->getGroupName())
                {
                    $this->returnStatus = Task::DELAYED;
                    return true;
                }
            }
        }

        // After all known tasks have run, check if all have ended without a CANCELLED status
        $taskStorage = $this->loadTasksLib()->getTaskStorage();
        $tasks = $taskStorage->getTasksByAttribute('groupName', $this->groupName);
        foreach ($tasks as $t)
        {
            if ($t->getStatus() === Task::CANCELLED)
            {
                $this->returnStatus = Task::CANCELLED;
                $task->setOutput('', 'Task cancelled due to failed group dependency.');
                return true;
            }
        }

        return false;
    }

    /**
     * @inheritDoc
     */
    public function blockCode(): int
    {
        return $this->returnStatus;
    }

    /**
     * @inheritDoc
     */
    public function delayTime(): int
    {
        return time() + 3;
    }

    /**
     * Load the tasks library, so that dependencies can get scanned later
     *
     * @return Tasks
     * @throws TasksException
     */
    private function loadTasksLib(): Tasks
    {
        try {
            /** @var Libraries $libraries */
            $libraries = Factory::getInstance('libraries');

            /** @var Tasks $tasks */
            $tasks = $libraries->get('async');

            return $tasks;
        } catch (FactoryException | LibraryException $e) {
            throw new TasksException("Could not constrain task. Async library could not be loaded.");
        }
    }
}